import faiss
import numpy as np
import json
from tqdm import tqdm
from time import time
from RAG_strategy import multi_query
from ragas import evaluate
from ragas.metrics import answer_relevancy, faithfulness, context_recall, context_precision
from datasets import Dataset
from typing import List, Callable

from dotenv import load_dotenv
load_dotenv('environment.env')
import os
import pickle
from supabase.client import Client, create_client
from langchain_openai import OpenAIEmbeddings
from add_vectordb import GetVectorStore
from sqlalchemy import create_engine
import pandas as pd
from langchain_core.documents import Document

# Load environment variables
supabase_url = os.getenv("SUPABASE_URL")
supabase_key = os.getenv("SUPABASE_KEY")
openai_api_key = os.getenv("OPENAI_API_KEY")
document_table = "documents"
URI = os.getenv("SUPABASE_URI")

# Initialize Supabase client
supabase: Client = create_client(supabase_url, supabase_key)

# Initialize embeddings
embeddings = OpenAIEmbeddings(openai_api_key=openai_api_key)

# Initialize original vector store
original_vector_store = GetVectorStore(embeddings, supabase, document_table)
original_retriever = original_vector_store.as_retriever(search_kwargs={"k": 4})



def download_embeddings():
    response = supabase.table(document_table).select("id, embedding, metadata, content").execute()
    embeddings = []
    ids = []
    metadatas = []
    contents = []
    for item in response.data:
        # Parse the embedding string into a list of floats
        embedding = json.loads(item['embedding'])
        embeddings.append(embedding)
        ids.append(item['id'])
        metadatas.append(item['metadata'])
        contents.append(item['content'])
    return np.array(embeddings, dtype=np.float32), ids, metadatas, contents

def create_faiss_index(embeddings):
    dimension = embeddings.shape[1]
    index = faiss.IndexFlatL2(dimension)
    index.add(embeddings)
    return index

def save_faiss_index(index, file_path):
    faiss.write_index(index, file_path)
    print(f"FAISS index saved to {file_path}")

def load_faiss_index(file_path):
    if os.path.exists(file_path):
        index = faiss.read_index(file_path)
        print(f"FAISS index loaded from {file_path}")
        return index
    return None

def save_metadata(ids, metadatas, contents, file_path):
    with open(file_path, 'wb') as f:
        pickle.dump((ids, metadatas, contents), f)
    print(f"Metadata saved to {file_path}")

def load_metadata(file_path):
    if os.path.exists(file_path):
        with open(file_path, 'rb') as f:
            ids, metadatas, contents = pickle.load(f)
        print(f"Metadata loaded from {file_path}")
        return ids, metadatas, contents
    return None, None, None

def search_faiss(index, query_vector, k=4):
    # Convert query_vector to a numpy array if it's not already
    if not isinstance(query_vector, np.ndarray):
        query_vector = np.array(query_vector)
    
    # Ensure the query_vector is 2D
    if query_vector.ndim == 1:
        query_vector = query_vector.reshape(1, -1)
    
    distances, indices = index.search(query_vector, k)
    return distances[0], indices[0]

class FAISSRetriever:
    def __init__(self, index, ids, metadatas, contents):
        self.index = index
        self.ids = ids
        self.metadatas = metadatas
        self.contents = contents
        self.embeddings_model = OpenAIEmbeddings(openai_api_key=openai_api_key)

    def get_relevant_documents(self, query: str) -> List[Document]:
        query_vector = self.embeddings_model.embed_query(query)
        _, indices = search_faiss(self.index, query_vector)
        return [
            Document(page_content=self.contents[i], metadata=self.metadatas[i])
            for i in indices
        ]

    def as_retriever(self, search_kwargs=None):
        return self

    def map(self) -> Callable[[List[str]], List[List[Document]]]:
        def _map(queries: List[str]) -> List[List[Document]]:
            return [self.get_relevant_documents(query) for query in queries]
        return _map

def load_qa_pairs():
    df = pd.read_csv("QA_database_rows.csv")
    return df['Question'].tolist(), df['Answer'].tolist()

if __name__ == "__main__":
    faiss_index_path = "faiss_index.bin"
    metadata_path = "faiss_metadata.pkl"

    # Try to load existing FAISS index and metadata
    index = load_faiss_index(faiss_index_path)
    ids, metadatas, contents = load_metadata(metadata_path)

    if index is None or ids is None:
        print("FAISS index or metadata not found. Creating new index...")
        print("Downloading embeddings from Supabase...")
        embeddings, ids, metadatas, contents = download_embeddings()

        print("Creating FAISS index...")
        index = create_faiss_index(embeddings)

        # Save the index and metadata
        save_faiss_index(index, faiss_index_path)
        save_metadata(ids, metadatas, contents, metadata_path)
    else:
        print("Using existing FAISS index and metadata.")

    print("Creating FAISS retriever...")
    faiss_retriever = FAISSRetriever(index, ids, metadatas, contents)

    # Load QA pairs from database
    questions, ground_truths = load_qa_pairs()

    # Compare performance
    for question, ground_truth in zip(questions, ground_truths):
        print(f"\nQuestion: {question}")

        # Measure time for FAISS retrieval
        start_time = time()
        faiss_answer, faiss_docs = multi_query(question, faiss_retriever, chat_history=[])
        faiss_time = time() - start_time
        print(f"FAISS Answer: {faiss_answer}")
        print(f"FAISS Time: {faiss_time:.4f} seconds")

        # Measure time for original retrieval
        start_time = time()
        original_answer, original_docs = multi_query(question, original_retriever, chat_history=[])
        original_time = time() - start_time
        print(f"Original Answer: {original_answer}")
        print(f"Original Time: {original_time:.4f} seconds")

        # RAGAS evaluation for FAISS
        faiss_datasets = {
            "question": [question],
            "answer": [faiss_answer],
            "contexts": [[doc.page_content for doc in faiss_docs]],
            "ground_truths": [[ground_truth]]
        }
        faiss_evalsets = Dataset.from_dict(faiss_datasets)

        faiss_result = evaluate(
            faiss_evalsets,
            metrics=[
                context_precision,
                faithfulness,
                answer_relevancy,
                context_recall,
            ],
        )

        print("FAISS RAGAS Evaluation:")
        print(faiss_result.to_pandas())

        # RAGAS evaluation for Original retriever
        original_datasets = {
            "question": [question],
            "answer": [original_answer],
            "contexts": [[doc.page_content for doc in original_docs]],
            "ground_truths": [[ground_truth]]
        }
        original_evalsets = Dataset.from_dict(original_datasets)

        original_result = evaluate(
            original_evalsets,
            metrics=[
                context_precision,
                faithfulness,
                answer_relevancy,
                context_recall,
            ],
        )

        print("Original RAGAS Evaluation:")
        print(original_result.to_pandas())

    print("\nPerformance comparison complete.")

# Key points:

# If both the index and metadata (ids, metadatas, contents) are found, we don't need to download embeddings or recreate the index. The FAISS index already contains the embeddings.
# The FAISSRetriever class doesn't need the raw embeddings. It uses the FAISS index for similarity search and the metadata (ids, metadatas, contents) for returning document information.